/* 
  EMO - an emotional robot simulation
  Based on the circumplex model of affect and a kNN algorithm
  Written by Markus Bindhammer January 2015
  Processing version 2.2.1
*/

import ddf.minim.*; // sound lib

Minim minim; // init sound lib
AudioPlayer bird; // plays singing birds
AudioPlayer noise; // plays awful noise

float current_x_position_A = 360; // current slider position for brightness
float current_x_position_P = 360; // current slider position for sound
float Activation; // activation variable
float Brightness; // brightness varibale 
float Pleasure; // pleasure variable
float Sound; // sound variable
PFont f; // text init

float head_angle = 0; // head alignment angle in degrees
float blinker = 0; // variable for eye blink
float angle_delta_min = 0.01; // activiation minimum
float angle_delta_high = 0.06; // activiation > 0
float angle_delta_max = 0.1; // activiation maximum
float angle_delta_neutral = 0.025; // activiation = 0
float angle_delta_low = 0.015; // activiation < 0
boolean blinkState = false; // blink state of eyes

final float r = 5; // radius of the circumplex model of affect

void setup() {
  size(610, 300); // screen size
  minim = new Minim(this);
  // load .wav from the data folder
  bird = minim.loadFile("bird.wav");
  noise = minim.loadFile("noise.wav");
  f = createFont("Ariel", 14, true); // create font for slider labeling
  smooth();
  frameRate(60);
}

void draw() {
  environment(); // environment in which the robot lives
  emotional_agent(Pleasure, Activation); // robot
}

//****************************************** FUNCTIONS ******************************************
// robot environment
void environment() {
  Brightness = map(current_x_position_A, 305, 575 ,0, 255); // map slider x position to a value between 0 and 255
  background(Brightness); // adjust brightness of background
  stroke(0); 
  strokeWeight(2);
  fill(255);
  rect(305, 20, 290, 60, 7); // brightness slide bar
  line(310, 50, 590, 50);
  rect(305, 100, 290, 60, 7); // sound slide bar
  line(310, 130, 590, 130);
  fill(0);
  textFont(f, 14); 
  textAlign(LEFT); // labeling brightness slide bar
  text("-5", 310, 74);
  textAlign(CENTER);
  text("0", 450, 74);
  textAlign(RIGHT);
  text("+5", 590, 74);
  textAlign(LEFT);
  text("BRIGHTNESS: ", 310, 35);
  Activation = map(current_x_position_A, 305, 575 ,-5, 5); // map slider position into activation range [-5,5]
  textAlign(CENTER);
  text(Activation, 450, 35); // display value
  textAlign(LEFT); // labeling sound slide bar
  text("-5", 310, 154);
  textAlign(CENTER);
  text("0", 450, 154);
  textAlign(RIGHT);
  text("+5", 590, 154);
  textAlign(LEFT);
  text("SOUND: ", 310, 115);
  Pleasure = map(current_x_position_P, 305, 575 ,-5, 5); // map slider position into pleasure range [-5,5]
  textAlign(CENTER);
  text(Pleasure, 450, 115); // display value
  if(Pleasure < 0) {
    noise.setGain(map(current_x_position_P, 575, 305, -60, 5));
  }
  if(Pleasure > 0) {
    bird.setGain(map(current_x_position_P, 305, 575, -60, 5));
  }
  if(Pleasure > 0 && bird.isPlaying() == false) {
    noise.pause();
    bird.loop(1);
  }
  if(Pleasure < 0 && noise.isPlaying() == false) {
    bird.pause();
    noise.loop(1);
  }
  if (mousePressed == true && (mouseButton == LEFT) && mouseY <= 65 && mouseY >= 30) {
    current_x_position_A = mouseX;
    if(current_x_position_A <= 305) current_x_position_A = 305;
    if(current_x_position_A >=575) current_x_position_A = 575;
    fill(255);
    rect(current_x_position_A, 40, 20, 20, 7);
  } else {
    fill(255);
    rect(current_x_position_A, 40, 20, 20, 7);
  }
  if (mousePressed == true && (mouseButton == LEFT) && mouseY <= 145 && mouseY >= 110) {
    current_x_position_P = mouseX;
    if(current_x_position_P <= 305) current_x_position_P = 305;
    if(current_x_position_P >=575) current_x_position_P = 575;
    fill(255);
    rect(current_x_position_P, 120, 20, 20, 7);
  } else {
    fill(255);
    rect(current_x_position_P, 120, 20, 20, 7);
  }
}
// draw face
void face(float R, float G, float B) {
  noStroke();
  fill(103, 58, 29);
  triangle(-100, -100, -110, 0, -50, 0); // hair
  triangle(100, -100, 50, 0, 110, 0); // hair
  fill(247, 196, 177);
  ellipse(-100, -80, 40, 40); // ear
  ellipse(100, -80, 40, 40); // ear
  fill(237, 171, 139);
  ellipse(-100, -80, 20, 20); // ear
  ellipse(100, -80, 20, 20); // ear
  fill(R, G, B); 
  ellipse(0, -100, 200, 200); // face
  fill(103, 58, 29);
  arc(0, -100, 200, 200, radians(180), radians(300), CHORD); // hair
  arc(0, -100, 200, 200, radians(220), radians(360), CHORD); // hair
  strokeWeight(1);
  stroke(162, 97, 77);
  line(0, -65, -10, -55); // nose
  line(0, -65, 10, -55); // nose
}
// functions for different mouths
void mouth_neutral() {
  noFill();
  strokeWeight(4);
  stroke(201, 102, 79);
  arc(0, -100, 130, 140, radians(60), radians(120)); // mouth
}
void mouth_angry() {
  noFill();
  strokeWeight(4);
  stroke(201, 102, 79);
  arc(0, 40, 130, 140, radians(250), radians(290)); // mouth
  stroke(103, 58, 29);
  line(-50, -120, -15, -100); // eye brow
  line(50, -120, 15, -100); // eye brow
}
void mouth_sad() {
  noFill();
  strokeWeight(4);
  stroke(201, 102, 79);
  arc(0, 40, 130, 140, radians(250), radians(290)); // mouth
}
void mouth_depressed() {
  noFill();
  strokeWeight(4);
  stroke(201, 102, 79);
  arc(0, 40, 130, 140, radians(240), radians(300)); // mouth
  stroke(103, 58, 29);
  arc(-40, -145, 60, 60, radians(40), radians(120)); // eye brow
  arc(40, -145, 60, 60, radians(60), radians(140)); // eye brow
}
void mouth_happy() {
  noFill();
  strokeWeight(4);
  stroke(201, 102, 79);
  fill(255);
  arc(0, -100, 110, 160, radians(50), radians(130), CHORD); // mouth
}
void mouth_bored() {
  noFill();
  strokeWeight(4);
  stroke(201, 102, 79);
  line(-10, -30, 10, -30); // face
}
// functions for different eyes
void eyes_neutral(float blink) {
  strokeWeight(2);
  stroke(255);
  fill(103, 58, 29);
  ellipse(-40, -90, 17, 27 + blink); // eye
  ellipse(40, -90, 17, 27 + blink); // eye 
}
void eyes_excited(float blink) {
  strokeWeight(2);
  stroke(255);
  fill(103, 58, 29);
  ellipse(-40, -90, 20, 20 + blink); // eye
  ellipse(40, -90, 20, 20 + blink); // eye 
}
void eyes_bored() {
  noFill();
  strokeWeight(4);
  stroke(103, 58, 29);
  arc(-40, -115, 60, 60, radians(40), radians(120)); // eye
  arc(40, -115, 60, 60, radians(60), radians(140)); // eye
}
/* 
  Emotion output function
  mood 0 = neutral, mood 1 = sad, mood 2 = happy, mood 3 = depressed
  mood 4 = bored, mood 5 = relaxed, mood 6 = angry, mood 7 = excited
  mood 8 = vigilant
*/
void emo(int mood) {
  float random_number = 0; // determines how often random blinks occur
  float random_select = 0;
  if(mood == 0 || mood == 1 || mood == 2) {
    head_angle += angle_delta_neutral;
    if (head_angle > 3 || head_angle < -3) {
      angle_delta_neutral = -angle_delta_neutral;
      head_angle += angle_delta_neutral;
    }
  }
  if(mood == 3 || mood == 5) {
    head_angle += angle_delta_low;
    if (head_angle > 3 || head_angle < -3) {
      angle_delta_low = -angle_delta_low;
      head_angle += angle_delta_low;
    }
  }
  if(mood == 4) {
    head_angle += angle_delta_min;
    if (head_angle > 3 || head_angle < -3) {
      angle_delta_min = -angle_delta_min;
      head_angle += angle_delta_min;
    }
  }
  if(mood == 6 || mood == 7) {
    head_angle += angle_delta_high;
    if (head_angle > 3 || head_angle < -3) {
      angle_delta_high = -angle_delta_high;
      head_angle += angle_delta_high;
    }
  }
  if(mood == 8) {
    head_angle += angle_delta_max;
    if (head_angle > 3 || head_angle < -3) {
      angle_delta_max = -angle_delta_max;
      head_angle += angle_delta_max;
    }
  }
  if(mood == 0 || mood == 1 || mood == 2) {
    random_number = 10000;
    random_select = 9900;
  }
  if(mood == 3) {
    random_number = 10000;
    random_select = 9990;
  }
  if(mood == 6 || mood == 7) {
    random_number = 10000;
    random_select = 9500;
  }
  if(mood == 8) {
    random_number = 10000;
    random_select = 9200;
  }
  if(random(0, random_number) > random_select && blinkState == false) blinkState = true; 
  if(blinkState == true) {
    blinker -= 2;
    pushMatrix();
    translate(150, 250); // center of rotation
    rotate(radians(head_angle)); 
    if(mood == 0) {
      face(247, 196, 177); // neutral face color
      mouth_neutral();
      eyes_neutral(blinker);
    }
    if(mood == 1) {
      face(247, 196, 177); // neutral face color
      mouth_sad();
      eyes_neutral(blinker);
    }
    if(mood == 2) {
      face(247, 196, 177); // neutral face color
      mouth_happy();
      eyes_neutral(blinker);
    }
    if(mood == 3) {
      face(248, 212, 201); // depressed face color
      mouth_depressed();
      eyes_neutral(blinker);
    }
    if(mood == 6) {
      face(242, 160, 130); // angry face color
      mouth_angry();
      eyes_neutral(blinker);
    }
    if(mood == 7) {
      face(242, 160, 130); // angry face color
      mouth_neutral();
      eyes_excited(blinker);
    }
    if(mood == 8) {
      face(239, 143, 107); // excited face color
      mouth_bored();
      eyes_excited(blinker);
    }
    popMatrix();
    if(blinker <= -22) {
      blinker = 0;
      blinkState = false;
    }  
  } else {
    pushMatrix();
    translate(150, 250); // center of rotation
    rotate(radians(head_angle)); 
    if(mood == 0) {
      face(247, 196, 177); // neutral face color
      mouth_neutral();
      eyes_neutral(0);
    }
    if(mood == 1) {
      face(247, 196, 177); // neutral face color
      mouth_sad();
      eyes_neutral(blinker);
    }
    if(mood == 2) {
      face(247, 196, 177); // neutral face color
      mouth_happy();
      eyes_neutral(blinker);
    }
    if(mood == 3) {
      face(248, 212, 201); // depressed face color
      mouth_depressed();
      eyes_neutral(blinker);
    }
    if(mood == 4) {
      face(251, 230, 223); // bored face color
      mouth_bored();
      eyes_bored();
    }
    if(mood == 5) {
      face(248, 212, 201); // depressed face color
      mouth_neutral();
      eyes_bored();
    }
    if(mood == 6) {
      face(242, 160, 130); // angry face color
      mouth_angry();
      eyes_neutral(blinker);
    }
    if(mood == 7) {
      face(242, 160, 130); // angry face color
      mouth_neutral();
      eyes_excited(blinker);
    }
    if(mood == 8) {
      face(239, 143, 107); // excited face color
      mouth_bored();
      eyes_excited(blinker);
    }
    popMatrix();
  }
}
// compute emotion
void emotional_agent(float x_P, float y_A) {
  float emotion_coordinates[][] = {
    {0, -r, r, -r/sqrt(2), 0, r/sqrt(2), -r/sqrt(2), r/sqrt(2), 0}, // x-coordinate of training samples
    {0, 0, 0, -r/sqrt(2), -r, -r/sqrt(2), r/sqrt(2), r/sqrt(2), r} // y- coordinate of training samples
  };
  byte i_emotion; 
  byte closestEmotion = 0;
  float MaxDiff;
  float MinDiff = sqrt(2) * r + r; // set MinDiff initially to maximum distance that can occure
  for (i_emotion = 0; i_emotion < 9; i_emotion ++) { 
    // compute Euclidean distances
    float Euclidian_distance = sqrt(pow((emotion_coordinates[0][i_emotion] - x_P), 2) + pow((emotion_coordinates[1][i_emotion] - y_A), 2));
    MaxDiff = Euclidian_distance;
    // find minimum distance 
    if (MaxDiff < MinDiff) {
      MinDiff = MaxDiff;
      closestEmotion = i_emotion;
    }
  }
  emo(closestEmotion);
}
